SanityItemCache = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0")
SanityItemCache:RegisterDB("SanityItemCacheDB", "SanityItemCacheDBPC")

local default_player_setup = {
	Characters = {},
	BankItems = {},
	BagItems = {},
	MailItems = {},
	EquipmentItems = {},
	Bags = {},
	Bank = {},
	Money = 0
}

local function translate_index(self, k)
	local keys = {["Count"] = 1, ["Bag"] = 2, ["Link"] = 3, ["Soulbound"] = 4, ["Slot"] = 5}
	
	-- We really don't want to blow our stack on valid numeric indexes
	if tonumber(k) then
		setmetatable(self,nil)
		local v = self[k]
		setmetatable(self,{__index = translate_index})
		return v		
	end
	
	if keys[k] then
		return self[keys[k]]
	elseif k ~= "Link" and self[3] and (
		k == "Name" or k == "ID" or k == "Quality" or k == "iLevel" or k == "MinLevel" or k == "Type" or
		k == "SubType" or k == "Stack" or k == "EquipLoc" or k == "Texture") then
		local n = GetItemInfo(self[3])
		if not n then
			self.invalid = true
		end
		local name, idstr, itemQuality, itemLevel, itemMinLevel, itemType, subType, stack, equipLoc, texture  = GetItemInfo(self[3]);
		if k == "Name" then
			self[k] = name or "Unknown"
			return self[k]
		elseif k == "ID" then
			self[k] = idstr or -1
			return self[k]
		elseif k == "Quality" then
			self[k] = itemQuality or 1
			return self[k]
		elseif k == "iLevel" then
			self[k] = itemLevel or 0
			return self[k]
		elseif k == "MinLevel" then
			self[k] = itemMinLevel or 0
			return self[k]
		elseif k == "Type" then
			self[k] = itemType or "Unknown"
			return self[k]
		elseif k == "SubType" then
			self[k] = subType or "Unknown"
			return self[k]
		elseif k == "Stack" then
			self[k] = stack or 1
			return self[k]
		elseif k == "EquipLoc" then
			self[k] = SANITY_CATEGORYNAMES[equipLoc] or equipLoc or "INVALID"
			return self[k]
		elseif k == "Texture" then
			self[k] = texture or "Interface\\Icons\\INV_Misc_QuestionMark"
			return self[k]
		end
	end
	return nil
end

SanityItemCache:RegisterDefaults("realm", { chars = {} })

function SanityItemCache:OnInitialize()
	self.char = UnitName("player")
	self:RegisterEvent("BAG_UPDATE")
	self:RegisterEvent("PLAYER_MONEY")
	self:RegisterEvent("BANKFRAME_OPENED")
	self:RegisterEvent("UNIT_INVENTORY_CHANGED")
	self:RegisterEvent("ADDON_LOADED")
	self:RegisterEvent("UPDATE_PENDING_MAIL")
	self:RegisterEvent("PLAYER_LOGOUT")
	if not self.db.realm.chars[self.char] then
		self.db.realm.chars[self.char] = {}
		for k,v in pairs(default_player_setup) do
			self.db.realm.chars[self.char][k] = {}
		end
	end
	self.INVENTORY_BAG_ID = -99
	self.MAILBOX_BAG_ID = -98
	SanityItemCache:PLAYER_MONEY()
end

function SanityItemCache:NukeCharacter(char)
	local tt = {}
	for tchar,dummy in pairs(self.db.realm.chars) do
		if tchar ~= char then
			tt[tchar] = dummy
		end
	end
	self.db.realm.chars = tt	
end

function SanityItemCache:ADDON_LOADED()
	local caches = {"BagItems", "BankItems", "EquipmentItems"}
	for i,cache in ipairs(caches) do
		for char,dummy in pairs(self.db.realm.chars) do
			for k,v in pairs(default_player_setup) do
				if type(self.db.realm.chars[char][k]) ~= type(v) and type(v) == "table" then
					self.db.realm.chars[char][k] = {}
				end
			end
			for k,v in pairs(self.db.realm.chars[char][cache]) do
				setmetatable(v,{__index = translate_index})
			end
		end
	end
end

-- This is basically a hack to keep Sanity from scanning a billion times on login
function SanityItemCache:UPDATE_PENDING_MAIL()
	if self.enable_collection then return end
	self.enable_collection = true
	SanityItemCache:BAG_UPDATE()
end

function SanityItemCache:PLAYER_LOGOUT()
	local caches = {"BagItems", "BankItems", "EquipmentItems"}
	for char,dummy in pairs(self.db.realm.chars) do
		for i,cache in ipairs(caches) do
			local t = {}
			for k,v in pairs(self.db.realm.chars[char][cache]) do
				tinsert(t, {
											[1] = v[1],
											[2] = v[2],
											[3] = v[3],
											[4] = v[4],
											[5] = v[5]
										})
			end
			self.db.realm.chars[char][cache] = t
		end
		local mt = {}
		for k,v in pairs(self.db.realm.chars[char].MailItems) do
			tinsert(mt, {
					["Name"] = v.Name,
					["Quality"] = v.Quality,
					["Count"] = v.Count,
					["Texture"] = v.Texture,
					["Index"] = v.Index,
					["Bag"] = v.Bag,
					["Description"] = v.Description
			})
		end
		self.db.realm.chars[char].MailItems = mt
	end
end

function SanityItemCache:BANKFRAME_OPENED(ignore)
	if ignore ~= "ignore" then
		self.db.realm.chars[self.char].BankItems, self.db.realm.chars[self.char].Bank = SanityItemCache:ScanBank()
	end
end

function SanityItemCache:UNIT_INVENTORY_CHANGED()
	SanityItemCache:BAG_UPDATE()
end

function SanityItemCache:BAG_UPDATE()
	if not self.enable_collection then return end
	self.db.realm.chars[self.char].BagItems, self.db.realm.chars[self.char].Bags = SanityItemCache:ScanBags()
	self.db.realm.chars[self.char].EquipmentItems = SanityItemCache:ScanEquipment()
	self:BANKFRAME_OPENED("ignore")
	Sanity.displayListDirty = true
	Sanity:Refresh()
end

function SanityItemCache:PLAYER_MONEY()
	self.db.realm.chars[self.char].Money = GetMoney()
end

function SanityItemCache:ScanBags()
	local bags = {0,1,2,3,4,KEYRING_CONTAINER}
	return SanityItemCache:Scan(bags)
end

function SanityItemCache:ScanBank()
	if Sanity.BankIsVisible then
		local bags = {11, 10, 9, 8, 7, 6, 5, -1}
		return SanityItemCache:Scan(bags)
	end
end

function SanityItemCache:Scan(bags)
	local items = {}
	local bagData = {bag_slots = 0, filled_slots = 0}
	for idx, bag in pairs(bags) do
		bagData.bag_slots = (bagData.bag_slots or 0) + GetContainerNumSlots(bag)
		for slot = 1, GetContainerNumSlots(bag) do
			local link = GetContainerItemLink(bag, slot)
			if link then
				bagData.filled_slots = (bagData.filled_slots or 0) + 1
				local n, _, itemQuality, itemLevel, itemMinLevel, itemType, subType, stack, equipLoc, texture  = GetItemInfo(link);
				if itemQuality then
					Sanity_ItemTooltip:ClearLines()
					if bag == -1 then
						Sanity_ItemTooltip:SetInventoryItem("player", BankButtonIDToInvSlotID(slot, nil));
					else
						Sanity_ItemTooltip:SetBagItem(bag, slot)
					end
	
					local soulbound = 0
					if bag ~= KEYRING_CONTAINER then
						for i=1,Sanity_ItemTooltip:NumLines() do
							local obj = getglobal("Sanity_ItemTooltipTextLeft" .. i)
							if obj:GetText() == SANITY_SOULBOUND or obj:GetText() == SANITY_QUESTITEM then
								soulbound = 1
							end
						end
					else
						soulbound = 1
					end
					
					local texture, count, locked = GetContainerItemInfo(bag, slot);
					local _,_,itemID = string.find(link, "|H(item:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+)")
					local item = {count, bag, itemID, soulbound, slot}
					setmetatable(item,{__index = translate_index})
					tinsert(items, item)
				end
			end
		end
	end
	return items, bagData
end

function SanityItemCache:ScanEquipment()
	local items = {}
	for i = 1,19 do
		Sanity_ItemTooltip:SetInventoryItem("player", i)
		local soulbound = 0
		for i=1,Sanity_ItemTooltip:NumLines() do
			if Sanity_ItemTooltipTextLeft2:GetText() == SANITY_SOULBOUND or Sanity_ItemTooltipTextLeft2:GetText() == SANITY_QUESTITEM then
				soulbound = 1
			end
		end
		local link = GetInventoryItemLink("player", i)
		if link then
			local name = GetItemInfo(link)
			local _,_,itemID = string.find(link, "|H(item:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+)")
			local item = {GetInventoryItemCount("player", i), SanityItemCache.INVENTORY_BAG_ID, itemID, soulbound, i}
			setmetatable(item,{__index = translate_index})
			tinsert(items, item)
		end
	end
	return items
end